/* Copyright (c) 1995-2001, VM Labs, Inc., All rights reserved.
 * Confidential and Proprietary Information of VM Labs, Inc
 * 1/25/01 kml
 *
 * NuonChannelManager is a base class intended to provide a
 * target-independent set of methods for initializing the system
 * for graphics, initializing channels, allocating pixmaps, etc.
 * The derived classes actually used will be for BIOS (Games), PE,
 * Windows emulation, and perhaps other targets.
 */

#ifndef NuonChannelManager_DEFINED
#define NuonChannelManager_DEFINED

#include <nuon/mml2d.h>
#include "NuonError.h"
#include "HSScan.h"
#include "NuonYccColorTable.h"

#ifndef WIN32
#include "auxvid.h"
#endif

enum {kChOsd = 0x1, kChMain = 0x2};
enum {kSysInitComplete = 0x01, 
	  	kOsdActive =       0x02, 
		kMainActive =   	 0x04,
		kOsdConfigured =   0x08,
		kMainConfigured =  0x10,
		kOsdVisible =      0x20,
		kMainVisible =     0x40
};

// note there are pure virtual functions in NuonChannelManager; it can only be used as
// a base class

class NuonChannelManager {

protected:
	unsigned mFlags;   // whether sys has been initialized, which channels are active, visible

public:
	// the constructor is where (for Nuon) we call mmlPowerUpGraphics()
	// something equivalent for Windows emulation
	NuonChannelManager();

	// accessor for system resources: actual data type returned is a function of the
	// platform, not always needed
	virtual const void *GetSystemResources (void) { return NULL; }

	// mmlInitGC(), something equiv for Windows 
	// the Graphics Context is needed for the mml2d rendering calls
	// I don't want to add another level of context-switching to the rendering calls
	// in NuonRaster::rasterize() by having them call something generic in this class
	// and calling the nuon-specific code from here
	// (adding this extra context-switch to initializing system/channel/pixmap kind of stuff 
	// just adds a little overhead once during the execution of the app)
	virtual int InitChannel (const unsigned channel = kChOsd) = 0;
	virtual void* GetLocalRamAdr (void) { return NULL; }

	// Show and Hide take an (optional) argument of channels
	// For Show and Hide, setting a channels flag for an inactive (uninitialized) channel 
	// is acknowledged in the return status but otherwise ignored as follows:
	// returns 0 on success, <0 on failure, >0 on one or both requested channels ignored
	// Config should check mmlDisplayPixmap::dmaFlags to verify pixmap can be displayed on
	// the selected channel
	
	// configure a channel before a call to Show()
	// this is where the check for allowed pixel format(s) is done
	virtual int ConfigChannel (mmlDisplayPixmap *sP, const unsigned channel = kChOsd) = 0;

	// show channels that have been configured with a call to ConfigChannel
	virtual int Show (const unsigned channels, const int alpha) = 0;

	// combines configure and show with calls to those functions, useful if only one channel is used
	virtual int Show (mmlDisplayPixmap *sP, const unsigned channel, const int alpha);

	// hide channels
	virtual int Hide (const unsigned channels = (kChOsd | kChMain));

	// surrender channels
	virtual int ReleaseChannel (const unsigned channels = (kChOsd | kChMain));

	// Loads the OSD hardware Clut
	virtual void SetClut (const NuonYccColorTable *clut) = 0;
	
	// Call with a new whichBuf to do a buffer flip (buffers all attached to the mmlDisplayPixmap
	// whichBuf must be less than the number of buffers allocated in the call to InitPixmaps
	virtual void SetDisplayFrame (mmlDisplayPixmap *sP, 
						 const unsigned channel = kChOsd, const unsigned whichBuf = 0) = 0;

	// Allocate pixmap(s)
	// returns mml status or Windows equiv (check return type)
	// uses the address specified in appMem or dispMem if not null
	
	// returns nonzero on failure, 0 on success
	// specifying depth defaults pixel format to RGBA
	virtual int InitPixmaps (mmlAppPixmap *sP, void *appMem, int wide, int high, 
						 int depth, int numBuffers = 1) = 0;

	virtual int InitPixmaps (mmlAppPixmap *sP, void *appMem, int wide, int high,
						 mmlPixFormat pix, int numBuffers = 1) = 0;

	// returns nonzero on failure, 0 on success
	// specifying depth defaults pixel format to YCCA
	virtual int InitPixmaps( mmlDisplayPixmap* sP, void *dispMem, int wide, int high, 
						 int depth, int numBuffers = 1) = 0;

	virtual int InitPixmaps( mmlDisplayPixmap* sP, void *dispMem, int wide, int high,
						 mmlPixFormat pix, int numBuffers = 1) = 0;

	// these do the same thing, but free the caller from needing to cast the Pixmap type
	virtual void ReleasePixmaps (mmlAppPixmap *sp, int numBuffers = 1) = 0;
	virtual void ReleasePixmaps (mmlDisplayPixmap *sp, int numBuffers = 1) = 0;

	// raster buffer operations
	virtual void CopyDisplay(mmlDisplayPixmap *sp, hsIntRect *rect, int x, int y, 
							const unsigned channel = kChOsd) = 0;
	virtual void CopyRect8 (mmlAppPixmap *appP, mmlDisplayPixmap *sP,
							const hsIntRect *rect, const unsigned channel = kChOsd) = 0;
	virtual void CopyRect16 (mmlAppPixmap *appP, mmlDisplayPixmap *sP,
							const hsIntRect *rect, const unsigned channel = kChOsd) = 0;
	virtual void CopyRect32 (mmlAppPixmap *appP, mmlDisplayPixmap *sP,
							const hsIntRect *rect, const unsigned channel = kChOsd) = 0;


#if defined (PRINT_DEBUG)
	virtual void PrintState (char *objectName);
#endif
};


// Nuon Channel Managers -- Move these to different header files as they get created
#if 0
// manager for the presentation engine
class NuonPeChannelManager: public NuonChannelManager {};

#endif

#endif  //NuonChannelManager_DEFINED



